#ifndef __dtnmc_pkt_h__
#define __dtnmc_pkt_h__

#define  DTNMCTYPE_PKT         0x01
#define  DTNMCTYPE_MCAST_PKT   0x02
#define  DTNMCTYPE_HELLO       0x03
#define  DTNMCTYPE_DELEGATE    0x04
#define  DTNMCTYPE_SUMVEC      0x05
#define  DTNMCTYPE_REQVEC      0x06

enum { NEW_PKT, DUPE_PKT };

#define  MAX_NEIGHBOR  5
#define  MAX_PKT_ACK   10 
#define  NEIGHBOR_END  (-1)

#define VECSIZE 200

#include <packet.h>

#define HDR_DTNMC_PKT(p) ((struct hdr_dtnmc_pkt *)hdr_dtnmc_pkt::access(p))

#define HDR_SUMVEC_PKT(p) ((struct hdr_sumvec_pkt *)hdr_sumvec_pkt::access(p))

struct hdr_dtnmc_pkt {
   nsaddr_t   pkt_src_;      // node which is sending (but not necessarily originating) this packet
   nsaddr_t   pkt_osrc_;     // node which originated this packet
   nsaddr_t   mcast_grp_id_; // multicast group id
   nsaddr_t   pkt_dest_;     // packet destination 
   u_int16_t  pkt_seq_num_;  // packet sequence number
   u_int8_t   dh_type_;      // packet type
   u_int32_t  no_copy_;      // number of copies (for spray and wait) 
   double     gen_time_;     // packet generation time 
   double     ttk_;          // time to kill (message expiration time) 
   double     member_since_; // starting time of membership of multicast group: mcast_grp_id_
   double     member_till_;  // stopping time of membership of multicast group: mcast_grp_id_
   bool       is_core_;      // flag to show if node is core

   int neighbors_[MAX_NEIGHBOR];  // neighbor information
   //struct pkt_ack pkt_acks_[MAX_NEIGHBOR];  // for packet acknowlegement  

   inline nsaddr_t&   pkt_src()       { return pkt_src_; }
   inline nsaddr_t&   pkt_osrc()      { return pkt_osrc_; }
   inline nsaddr_t&   pkt_dest()      { return pkt_dest_; }
   inline nsaddr_t&   mcast_grp_id()  { return mcast_grp_id_; }
   inline u_int16_t&  pkt_seq_num()   { return pkt_seq_num_; }
   inline u_int8_t&   dh_type()       { return dh_type_; }
   inline u_int32_t&  no_copy()       { return no_copy_; }
   inline double&     gen_time()      { return gen_time_; }
   inline double&     ttk()           { return ttk_; }
   inline double&     member_since()  { return member_since_; }
   inline double&     member_till()   { return member_till_; }
   inline bool&       is_core()       { return is_core_; }

   inline int&        neighbor(int i) { 
      assert(i < MAX_NEIGHBOR);
      return neighbors_[i];
   }

   static int offset_;

   inline static int& offset() { return offset_ ;}

   inline static hdr_dtnmc_pkt* access(const Packet* p) {
      return (hdr_dtnmc_pkt*) p->access(offset_);
   }

   inline int size() {
      int sz = 0;

      sz = sizeof(nsaddr_t) +    // pkt_osrc_
           sizeof(nsaddr_t) +    // pkt_dest_
           sizeof(nsaddr_t) +    // pkt_src_
           sizeof(nsaddr_t) +    // pkt_grp_id_
           sizeof(double)   +    // gen_time_ 
           sizeof(double)   +    // ttk_ 
           sizeof(double)   +    // member_since_
           sizeof(double)   +    // member_till_
           sizeof(u_int16_t)+    // pkt_seq_num_
           sizeof(u_int8_t) +    // dh_type_
           sizeof(u_int32_t) +   // no_copy_
           sizeof(neighbors_)+   // neighbors (integer array)
           sizeof(u_int32_t)*2 ; // reserved

      assert(sz > 0);
      return sz;
   }
};

typedef struct {
   nsaddr_t  src;
   u_int16_t seqno;
} vecent ;

struct hdr_sumvec_pkt {
   nsaddr_t   pkt_src_;       // sending node 
   nsaddr_t   pkt_dest_;      // receiving node 
   nsaddr_t   mcast_grp_id_;  // multicast group id
   u_int32_t  vsize_;
   double     sum_since_;
   vecent vecs_[VECSIZE];
   u_int8_t   dh_type_;      // packet type

   inline nsaddr_t&   pkt_src()       { return pkt_src_; }
   inline nsaddr_t&   pkt_dest()      { return pkt_dest_; }
   inline nsaddr_t&   mcast_grp_id()  { return mcast_grp_id_; }
   inline double&     sum_since()     { return sum_since_; }
   inline u_int8_t&   dh_type()       { return dh_type_; }
   inline u_int32_t&  vsize()         { return vsize_; }

   static int offset_;

   inline static int& offset() { return offset_ ;}

   inline static hdr_sumvec_pkt* access(const Packet* p) {
      return (hdr_sumvec_pkt*) p->access(offset_);
   }

   inline vecent *vecs(int i) { 
      assert(i < VECSIZE);
      return &vecs_[i];
   }

   inline int size() {
      int sz = 0;

      sz = sizeof(nsaddr_t) +    // pkt_src_
           sizeof(nsaddr_t) +    // pkt_dest_ 
           sizeof(nsaddr_t) +    // pkt_grp_id_
           sizeof(double)   +    // sum_since_
           sizeof(vecs_[0]) * VECSIZE + // vecs (summary vector)
           sizeof(u_int32_t)*2 ; // reserved

      assert(sz > 0);
      return sz;
   }
};

struct cores {
   nsaddr_t core_id;
   struct cores *next;
};

struct dtnmc_pkt_info {
   nsaddr_t  src;  // previous src (not necessarily original sender)
   nsaddr_t  osrc;  // original src
   nsaddr_t  dest;  // final destination
   nsaddr_t  nhop;  // next hop 
   u_int16_t seqno; //sequence number of the pkt. tuple (osrc, dest, seqno) uniquely determines a pkt
   u_int32_t nocopy; // number of copies delegated (for spray and wait)
   u_int8_t  dhtype;
   int       ttl;           // time to live (hop count)
   double    ttk;           // time to kill (message expiration time) 
   nsaddr_t  gid;           // multicast group id
   double    gen_time;      // packet generation time 
   double    member_since;  // start of multicast membership interval
   double    member_till;   // end of multicast membership interval
   double    time_stored;   // time the packet is stored in the buffer
   struct cores *dlg_cores; // keep track of core that have delegated the packet
};

/* To decrease number of redundent message transfers, we use packet acknowledgement.
 * We use this for gossip based spray and wait. 
 */
struct pkt_ack {
   nsaddr_t osrc;
   nsaddr_t dest;
   nsaddr_t gid;
   u_int16_t seqno; //pkt sequence number. tuple (osrc, dest/gid, seqno) uniquely determines a pkt
   int status; // shows if this packet has been delivered before
};

#endif
